"""Test realistic scenarios, which are slower."""

import pytest
import xarray as xr

import particle_tracking_manager as ptm


@pytest.mark.slow
def test_add_new_reader():
    """Add a separate reader from the defaults."""

    import xroms

    manager = ptm.OpenDriftModel(use_static_masks=True, steps=2)

    url = xroms.datasets.CLOVER.fetch("ROMS_example_full_grid.nc")
    ds = xr.open_dataset(url, decode_times=False)
    manager.add_reader(ds=ds)


@pytest.mark.slow
def test_run():
    """Set up and run."""

    import xroms

    seeding_kwargs = dict(lon=-90, lat=28.7, number=1)
    manager = ptm.OpenDriftModel(**seeding_kwargs, use_static_masks=True, steps=2)
    url = xroms.datasets.CLOVER.fetch("ROMS_example_full_grid.nc")
    ds = xr.open_dataset(url, decode_times=False)
    manager.add_reader(ds=ds, name="txla")
    manager.seed()
    manager.run()



"""Test configuration details."""

import particle_tracking_manager as ptm


def test_ptm_config():
    """Test that PTM config is brought in

    ...and takes precedence over model config.
    """

    m = ptm.OpenDriftModel()

    # check for a single key and make sure override default is present
    assert m.show_config()["coastline_action"]["default"] == "previous"


def test_show_config():
    """Test configuration-showing functionality."""

    m = ptm.OpenDriftModel()

    # check sorting by a single level
    assert sorted(m.show_config(level=1).keys()) == [
        "seed:seafloor",
        "seed:z",
        "seed_seafloor",
        "z",
    ]

    # check PTM level sorting
    assert (
        "lon" in m.show_config(ptm_level=1).keys()
        and "log" not in m.show_config(ptm_level=1).keys()
    )


def test_surface_only():
    """Make sure appropriate parameters are set if surface_only is True."""

    m = ptm.OpenDriftModel(surface_only=True, z=-10)


def test_default_overrides():
    """Make sure input to OpenDriftModel and to PTM are represented as values in config."""

    m = ptm.OpenDriftModel(emulsification=False, drift_model="OpenOil", steps=5)

    # the value should be False due to input value
    # check without running update_config which since want to know initial state
    assert not m.show_config_model("processes:emulsification")["value"]
    assert (
        m._config["steps"]["value"]
        == m.config_ptm["steps"]["value"]
        == m.show_config_model("steps")["value"]
        == 5
    )



"""Command line interface to get inputs from web application."""

import argparse
import ast
import logging

from datetime import datetime

import pandas as pd

import particle_tracking_manager as ptm


def is_int(s):
    """Check if string is actually int."""
    try:
        int(s)
        return True
    except (ValueError, TypeError):
        return False


def is_float(s):
    """Check if string is actually float."""
    try:
        float(s)
        return True
    except (ValueError, TypeError):
        return False


def is_None(s):
    """Check if string is actually None."""
    if s == "None":
        return True
    else:
        return False


def is_datestr(s):
    """Check if string is actually a datestring."""

    try:
        out = pd.Timestamp(s)
        assert not pd.isnull(out)
        return True
    except (ValueError, AssertionError):
        return False


def is_deltastr(s):
    """Check if string is actually a Timedelta."""

    try:
        out = pd.Timedelta(s)
        assert not pd.isnull(out)
        return True
    except (ValueError, AssertionError):
        return False


# https://sumit-ghosh.com/articles/parsing-dictionary-key-value-pairs-kwargs-argparse-python/
class ParseKwargs(argparse.Action):
    """With can user can input dicts on CLI."""

    def __call__(self, parser, namespace, values, option_string=None):
        """With can user can input dicts on CLI."""
        setattr(namespace, self.dest, dict())
        for value in values:
            # maxsplit helps in case righthand side of input has = in it, like filenames can have
            key, value = value.split("=", maxsplit=1)
            # catch list case
            if value.startswith("[") and value.endswith("]"):
                # if "[" in value and "]" in value:
                value = value.strip("][").split(",")
            # change numbers to numbers but with attention to decimals and negative numbers
            if is_int(value):
                value = int(value)
            elif is_float(value):
                value = float(value)
            elif is_None(value):
                value = None
            elif is_deltastr(value):
                value = pd.Timedelta(value)
            elif is_datestr(value):
                value = pd.Timestamp(value)
            getattr(namespace, self.dest)[key] = value


def main():
    """Parser method.

    Include all inputs

    Example
    -------

    >>> python cli.py lon=-151 lat=59 use_auto_landmask=True start_time=2000-1-1
    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "kwargs",
        nargs="*",
        action=ParseKwargs,
        help="Input keyword arguments for running PTM. Available options are specific to the `catalog_type`. Dictionary-style input, e.g. `case='Oil'`. Format for list items is e.g. standard_names='[sea_water_practical_salinity,sea_water_temperature]'.",
    )

    parser.add_argument(
        "--dry-run",
        help="Return configuration parameters without running the model.",
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    args = parser.parse_args()

    to_bool = {
        key: ast.literal_eval(value)
        for key, value in args.kwargs.items()
        if value in ["True", "False"]
    }
    args.kwargs.update(to_bool)

    if "output_file" not in args.kwargs:

        args.kwargs[
            "output_file"
        ] = f"output-results_{datetime.utcnow():%Y-%m-%dT%H%M:%SZ}.nc"

    log_file = args.kwargs["output_file"].replace(".nc", ".log")

    # Create a file handler
    file_handler = logging.FileHandler(log_file)

    # Create a formatter and add it to the handler
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)

    m = ptm.OpenDriftModel(**args.kwargs)

    if args.dry_run:

        # run this to make sure everything is updated fully
        m.add_reader()
        print(m.drift_model_config())

    else:

        # Add the handler to the logger
        m.logger.addHandler(file_handler)

        m.logger.info(f"filename: {args.kwargs['output_file']}")

        m.add_reader()
        print(m.drift_model_config())

        m.seed()
        m.run()

        print(m.outfile_name)

    # Remove the handler at the end of the loop
    m.logger.removeHandler(file_handler)
    file_handler.close()



"""Test manager use in library, the default approach."""

import unittest

from datetime import datetime, timedelta
from unittest import mock

import numpy as np
import pandas as pd
import pytest

import particle_tracking_manager as ptm


def test_order():
    """Have to configure before seeding."""

    with pytest.raises(KeyError):
        manager = ptm.OpenDriftModel()
        manager.run()


def test_seed():
    """make sure seeding works with no ocean model

    also compare two approaches for inputting info.
    """

    manager = ptm.OpenDriftModel(use_auto_landmask=True, number=1)
    manager.lon = -151
    manager.lat = 59
    manager.start_time = datetime(2000, 1, 1)

    # with pytest.raises(ValueError):
    #     manager.seed()

    manager.ocean_model = "test"
    manager.has_added_reader = True  # cheat to run test
    manager.seed()
    # look at elements with manager.o.elements_scheduled

    seeding_kwargs = dict(lon=-151, lat=59, start_time=datetime(2000, 1, 1))
    manager2 = ptm.OpenDriftModel(
        use_auto_landmask=True, number=1, ocean_model="test", **seeding_kwargs
    )
    manager2.has_added_reader = True  # cheat to run test
    manager2.seed()

    assert (
        manager.o.elements_scheduled.__dict__ == manager2.o.elements_scheduled.__dict__
    )


@mock.patch(
    "particle_tracking_manager.models.opendrift.opendrift.OpenDriftModel.reader_metadata"
)
def test_lon_check(mock_reader_metadata):
    """Test longitude check that is run when variable and reader are set."""

    # Check that longitude is checked as being within (mocked) reader values
    mock_reader_metadata.return_value = np.array([-150, -140, -130])

    m = ptm.OpenDriftModel(lon=0, lat=0)

    # this causes the check
    with pytest.raises(AssertionError):
        m.has_added_reader = True


@mock.patch(
    "particle_tracking_manager.models.opendrift.opendrift.OpenDriftModel.reader_metadata"
)
def test_start_time_check(mock_reader_metadata):
    """Test start_time check that is run when variable and reader are set."""

    # Check that start_time is checked as being within (mocked) reader values
    mock_reader_metadata.return_value = datetime(2000, 1, 1)

    m = ptm.OpenDriftModel(start_time=datetime(1999, 1, 1))

    # this causes the check
    with pytest.raises(AssertionError):
        m.has_added_reader = True


@mock.patch(
    "particle_tracking_manager.models.opendrift.opendrift.OpenDriftModel.reader_metadata"
)
def test_ocean_model_not_None(mock_reader_metadata):
    """Test that ocean_model can't be None."""

    # Use this to get through steps necessary for the test
    mock_reader_metadata.return_value = datetime(2000, 1, 1)

    m = ptm.OpenDriftModel()
    with pytest.raises(AssertionError):
        m.has_added_reader = True


@pytest.mark.slow
def test_parameter_passing():
    """make sure parameters passed into package make it to simulation runtime."""

    ts = 5
    diffmodel = "windspeed_Sundby1983"
    use_auto_landmask = True
    vertical_mixing = True
    do3D = True

    seed_kws = dict(
        lon=4.0,
        lat=60.0,
        radius=5000,
        number=100,
        start_time=datetime(2015, 9, 22, 6, 0, 0),
    )
    m = ptm.OpenDriftModel(
        use_auto_landmask=use_auto_landmask,
        time_step=ts,
        duration=timedelta(hours=10),
        steps=None,
        diffusivitymodel=diffmodel,
        vertical_mixing=vertical_mixing,
        do3D=do3D,
        **seed_kws
    )

    # idealized simulation, provide a fake current
    m.o.set_config("environment:fallback:y_sea_water_velocity", 1)

    # seed
    m.seed()

    # run simulation
    m.run()

    # check time_step across access points
    assert (
        m.o.time_step.seconds
        == ts
        == m.time_step
        == m.show_config_model(key="time_step")["value"]
    )

    # check diff model
    assert m.show_config(key="diffusivitymodel")["value"] == diffmodel

    # check use_auto_landmask coming through
    assert m.show_config(key="use_auto_landmask")["value"] == use_auto_landmask


def test_keyword_parameters():
    """Make sure unknown parameters are not input."""

    with pytest.raises(KeyError):
        m = ptm.OpenDriftModel(incorrect_key="test")


def test_ocean_model_timing():
    """Check that error raised when timing for model wrong...

    and not other times."""

    with pytest.raises(AssertionError):
        m = ptm.OpenDriftModel(ocean_model="NWGOA", start_time="1998-1-1")

    with pytest.raises(AssertionError):
        m = ptm.OpenDriftModel(ocean_model="NWGOA", start_time="2009-2-1")

    m = ptm.OpenDriftModel(ocean_model="NWGOA", start_time="2007-2-1")

    with pytest.raises(AssertionError):
        m = ptm.OpenDriftModel(ocean_model="CIOFS", start_time="1998-1-1")

    with pytest.raises(AssertionError):
        m = ptm.OpenDriftModel(ocean_model="CIOFS", start_time="2023-2-1")

    m = ptm.OpenDriftModel(ocean_model="CIOFS", start_time="2020-2-1")

    with pytest.raises(AssertionError):
        m = ptm.OpenDriftModel(ocean_model="CIOFSOP", start_time="2020-1-1")

    m = ptm.OpenDriftModel(ocean_model="CIOFSOP", start_time="2023-1-1")


def test_lon_lat_checks():
    """Check that lon/lat check errors are raised."""

    with pytest.raises(AssertionError):
        m = ptm.OpenDriftModel(lon=-180.1)

    m = ptm.OpenDriftModel(lon=0)

    with pytest.raises(AssertionError):
        m = ptm.OpenDriftModel(lat=-95)

    m = ptm.OpenDriftModel(lat=0)


def test_setattr_oceanmodel_lon0_360():
    """Test setting oceanmodel_lon0_360 attribute."""
    manager = ptm.OpenDriftModel()
    manager.lon = -150
    manager.oceanmodel_lon0_360 = True
    assert manager.lon == 210


def test_setattr_surface_only():
    """Test setting surface_only attribute."""
    manager = ptm.OpenDriftModel(do3D=True, z=1, vertical_mixing=True)
    manager.surface_only = True
    assert manager.do3D == False
    assert manager.z == 0
    assert manager.vertical_mixing == False


def test_input_too_many_end_of_simulation():
    with pytest.raises(AssertionError):
        ptm.OpenDriftModel(
            steps=4,
            duration=pd.Timedelta("24h"),
            end_time=pd.Timestamp("1970-01-01T02:00"),
        )


def test_changing_end_of_simulation():
    """change end_time, steps, and duration

    and make sure others are updated accordingly.
    This accounts for the default time_step of 300 seconds.

    """

    m = ptm.OpenDriftModel(start_time=pd.Timestamp("2000-1-1"))
    m.start_time = pd.Timestamp("2000-1-2")
    m.end_time = pd.Timestamp("2000-1-3")
    assert m.steps == 288
    assert m.duration == pd.Timedelta("1 days 00:00:00")

    m.steps = 48
    assert m.end_time == pd.Timestamp("2000-01-02 04:00:00")
    assert m.duration == pd.Timedelta("0 days 04:00:00")

    m.duration = pd.Timedelta("2 days 12:00:00")
    assert m.end_time == pd.Timestamp("2000-01-04 12:00:00")
    assert m.steps == 720


class TestTheManager(unittest.TestCase):
    def setUp(self):
        self.m = ptm.OpenDriftModel()
        self.m.reader_metadata = mock.MagicMock(
            side_effect=lambda x: {
                "lon": np.array([0, 180]),
                "lat": np.array([-90, 90]),
                "start_time": pd.Timestamp("2022-01-01 12:00:00"),
            }[x]
        )

    def test_has_added_reader_true_lon_lat_set(self):
        self.m.lon = 90
        self.m.lat = 45
        self.m.ocean_model = "test"
        self.m.has_added_reader = True
        self.assertEqual(self.m.has_added_reader, True)

    def test_has_added_reader_true_start_time_set(self):
        self.m.start_time = "2022-01-01 12:00:00"
        self.m.ocean_model = "test"
        self.m.has_added_reader = True
        self.assertEqual(self.m.has_added_reader, True)


class TestManager(unittest.TestCase):
    def setUp(self):
        self.m = ptm.OpenDriftModel()

    def test_start_time_str(self):
        self.m.start_time = "2022-01-01 12:00:00"
        self.assertEqual(self.m.start_time, pd.Timestamp("2022-01-01 12:00:00"))

    def test_start_time_datetime(self):
        dt = datetime(2022, 1, 1, 12, 0, 0)
        self.m.start_time = dt
        self.assertEqual(self.m.start_time, pd.Timestamp(dt))

    def test_start_time_timestamp(self):
        ts = pd.Timestamp("2022-01-01 12:00:00")
        self.m.start_time = ts
        self.assertEqual(self.m.start_time, ts)

    def test_start_time_invalid(self):
        with self.assertRaises(TypeError):
            self.m.start_time = 123

    def test_surface_only_true(self):
        self.m.surface_only = True
        self.m.do3D = True
        self.assertEqual(self.m.do3D, False)
        self.m.z = 10
        self.assertEqual(self.m.z, 0)
        self.m.vertical_mixing = True
        self.assertEqual(self.m.vertical_mixing, False)

    def test_surface_only_false_do3D_false(self):
        self.m.surface_only = False
        self.m.do3D = False
        self.m.vertical_mixing = True
        self.assertEqual(self.m.vertical_mixing, False)

    def test_surface_only_false_do3D_true(self):
        self.m.surface_only = False
        self.m.do3D = True
        self.m.vertical_mixing = True
        self.assertEqual(self.m.vertical_mixing, True)

    def test_seed_seafloor_true(self):
        self.m.seed_seafloor = True
        self.assertIsNone(self.m.z)

    def test_z_set(self):
        self.m.z = 10
        self.assertEqual(self.m.z, 10)
        self.assertFalse(self.m.seed_seafloor)

    def test_has_added_reader_true_ocean_model_set(self):
        self.m.ocean_model = "test"
        self.m.has_added_reader = True
        self.assertEqual(self.m.has_added_reader, True)

    def test_run_forward_true(self):
        self.m.run_forward = True
        self.assertEqual(self.m.timedir, 1)

    def test_run_forward_false(self):
        self.m.run_forward = False
        self.assertEqual(self.m.timedir, -1)

    def test_seed_flag_elements_lon_lat_none(self):
        self.m.seed_flag = "elements"
        self.m.lon = None
        self.m.lat = None
        with pytest.raises(KeyError):
            self.m.seed()

    def test_seed_flag_geojson_geojson_none(self):
        self.m.seed_flag = "geojson"
        self.m.geojson = None
        with pytest.raises(KeyError):
            self.m.seed()

    def test_seed_seafloor_false_z_none(self):
        self.m.seed_seafloor = False
        self.m.lon = 0
        self.m.lat = 0
        self.m.z = None
        with pytest.raises(AssertionError):
            self.m.seed()

    def test_start_time_none(self):
        self.m.start_time = None
        with pytest.raises(KeyError):
            self.m.seed()


if __name__ == "__main__":
    unittest.main()



"""Test seeding functionality."""

from datetime import datetime

import numpy as np
import pytest

import particle_tracking_manager as ptm


# def test_seeding_from_wkt():
#     """Seed from WKT."""

#     wkt = "POLYGON((-151.0 58.0, -150.0 58.0, -150.0 59.0, -151.0 59.0, -151.0 58.0))"

#     m = ptm.OpenDriftModel(seed_flag="wkt", wkt=wkt, use_auto_landmask=True)
#     m.seed()
#     import pdb; pdb.set_trace()


def test_seeding_from_geojson():
    """Seed from GeoJSON."""

    geo = {
        "type": "Feature",
        "properties": {},
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [-151.0, 58.0],
                    [-150.0, 58.0],
                    [-150.0, 59.0],
                    [-151.0, 59.0],
                    [-151.0, 58.0],
                ]
            ],
        },
    }
    m = ptm.OpenDriftModel(
        seed_flag="geojson",
        start_time="2000-01-01",
        geojson=geo,
        use_auto_landmask=True,
        number=2,
    )
    m.seed()

    expected_lon = [-150.51787, -150.51787]
    expected_lat = [58.25654, 58.25654]

    assert np.allclose(m.initial_drifters.lon, expected_lon)
    assert np.allclose(m.initial_drifters.lat, expected_lat)



"""Using OpenDrift for particle tracking."""
import copy
import datetime
import gc
import json
import logging
import os
import platform
import tempfile

from pathlib import Path
from typing import Optional, Union

import appdirs

# using my own version of ROMS reader
# from .reader_ROMS_native import Reader
import pandas as pd
import xarray as xr

from opendrift.models.larvalfish import LarvalFish
from opendrift.models.leeway import Leeway
from opendrift.models.oceandrift import OceanDrift
from opendrift.models.openoil import OpenOil
from opendrift.readers import reader_ROMS_native

from ...cli import is_None
from ...the_manager import _KNOWN_MODELS, ParticleTrackingManager


# from .cli import is_None
# from .the_manager import ParticleTrackingManager


# Read OpenDrift configuration information
loc = Path(__file__).parent / Path("config.json")
with open(loc, "r") as f:
    # Load the JSON file into a Python object
    config_model = json.load(f)

# convert "None"s to Nones
for key in config_model.keys():
    if "default" in config_model[key] and is_None(config_model[key]["default"]):
        config_model[key]["default"] = None


# logger = logging.getLogger("opendrift")


# @copydocstring( ParticleTrackingManager )
class OpenDriftModel(ParticleTrackingManager):
    """Open drift particle tracking model.

    Defaults all come from config_model configuration file.

    Parameters
    ----------
    drift_model : str, optional
        Options: "OceanDrift", "LarvalFish", "OpenOil", "Leeway", by default "OceanDrift"
    export_variables : list, optional
        List of variables to export, by default None. See PTM docs for options.
    radius : int, optional
        Radius around each lon-lat pair, within which particles will be randomly seeded. This is used by function `seed_elements`.
    radius_type : str
        If 'gaussian' (default), the radius is the standard deviation in x-y-directions. If 'uniform', elements are spread evenly and always inside a circle with the given radius. This is used by function `seed_elements`.

    horizontal_diffusivity : float
        Horizontal diffusivity is None by default but will be set to a grid-dependent value for known ocean_model values. This is calculated as 0.1 m/s sub-gridscale velocity that is missing from the model output and multiplied by an estimate of the horizontal grid resolution. This leads to a larger value for NWGOA which has a larger value for mean horizontal grid resolution (lower resolution). If the user inputs their own ocean_model information, they can input their own horizontal_diffusivity value. A user can use a known ocean_model and then overwrite the horizontal_diffusivity value to some value.
    current_uncertainty : float
        Add gaussian perturbation with this standard deviation to current components at each time step.
    wind_uncertainty : float
        Add gaussian perturbation with this standard deviation to wind components at each time step.
    use_auto_landmask : bool
        Set as True to use general landmask instead of that from ocean_model.
        Use for testing primarily. Default is False.
    diffusivitymodel : str
        Algorithm/source used for profile of vertical diffusivity. Environment means that diffusivity is acquired from readers or environment constants/fallback. Turned on if ``vertical_mixing==True``.
    stokes_drift : bool, optional
        Set to True to turn on Stokes drift, by default True. This enables 3 settings in OpenDrift:

        * o.set_config('drift:use_tabularised_stokes_drift', True)
        * o.set_config('drift:tabularised_stokes_drift_fetch', '25000')  # default
        * o.set_config('drift:stokes_drift_profile', 'Phillips')  # default

        The latter two configurations are not additionally set in OpenDriftModel since they are already the default once stokes_drift is True.
    mixed_layer_depth : float
        Fallback value for ocean_mixed_layer_thickness if not available from any reader. This is used in the calculation of vertical diffusivity.
    coastline_action : str, optional
        Action to perform if a drifter hits the coastline, by default "previous". Options
        are 'stranding', 'previous'.
    seafloor_action : str, optional
        Action to perform if a drifter hits the seafloor, by default "deactivate". Options
        are 'deactivate', 'previous', 'lift_to_seafloor'.
    max_speed : int
        Typical maximum speed of elements, used to estimate reader buffer size.
    wind_drift_factor : float
        Elements at surface are moved with this fraction of the wind vector, in addition to currents and Stokes drift.
    wind_drift_depth : float
        The direct wind drift (windage) is linearly decreasing from the surface value (wind_drift_factor) until 0 at this depth.
    vertical_mixing_timestep : float
        Time step used for inner loop of vertical mixing.
    object_type: str = config_model["object_type"]["default"],
        Leeway object category for this simulation.

    diameter : float
        Seeding value of diameter.
    neutral_buoyancy_salinity : float
        Seeding value of neutral_buoyancy_salinity.
    stage_fraction : float
        Seeding value of stage_fraction.
    hatched : float
        Seeding value of hatched.
    length : float
        Seeding value of length.
    weight : float
        Seeding value of weight.

    oil_type : str
        Oil type to be used for the simulation, from the NOAA ADIOS database.
    m3_per_hour : float
        The amount (volume) of oil released per hour (or total amount if release is instantaneous).
    oil_film_thickness : float
        Seeding value of oil_film_thickness.
    droplet_size_distribution : str
        Droplet size distribution used for subsea release.
    droplet_diameter_mu : float
        The mean diameter of oil droplet for a subsea release, used in normal/lognormal distributions.
    droplet_diameter_sigma : float
        The standard deviation in diameter of oil droplet for a subsea release, used in normal/lognormal distributions.
    droplet_diameter_min_subsea : float
        The minimum diameter of oil droplet for a subsea release, used in uniform distribution.
    droplet_diameter_max_subsea : float
        The maximum diameter of oil droplet for a subsea release, used in uniform distribution.
    emulsification : bool
        Surface oil is emulsified, i.e. water droplets are mixed into oil due to wave mixing, with resulting increase of viscosity.
    dispersion : bool
        Oil is removed from simulation (dispersed), if entrained as very small droplets.
    evaporation : bool
        Surface oil is evaporated.
    update_oilfilm_thickness : bool
        Oil film thickness is calculated at each time step. The alternative is that oil film thickness is kept constant with value provided at seeding.
    biodegradation : bool
        Oil mass is biodegraded (eaten by bacteria).
    log : str, optional
        Options are "low" and "high" verbosity for log, by default "low"

    Notes
    -----
    Docs available for more initialization options with ``ptm.ParticleTrackingManager?``

    """

    logger: logging.Logger
    log: str
    loglevel: str
    vertical_mixing_timestep: float
    diffusivitymodel: str
    mixed_layer_depth: float
    wind_drift_factor: float
    wind_drift_depth: float
    stokes_drift: bool
    drift_model: str
    o: Union[OceanDrift, Leeway, LarvalFish, OpenOil]
    horizontal_diffusivity: Optional[float]
    config_model: dict

    def __init__(
        self,
        drift_model: str = config_model["drift_model"]["default"],
        export_variables: str = config_model["export_variables"]["default"],
        radius: int = config_model["radius"]["default"],
        radius_type: str = config_model["radius_type"]["default"],
        horizontal_diffusivity: float = config_model["horizontal_diffusivity"][
            "default"
        ],
        current_uncertainty: float = config_model["current_uncertainty"]["default"],
        wind_uncertainty: float = config_model["wind_uncertainty"]["default"],
        use_auto_landmask: bool = config_model["use_auto_landmask"]["default"],
        diffusivitymodel: str = config_model["diffusivitymodel"]["default"],
        stokes_drift: bool = config_model["stokes_drift"]["default"],
        mixed_layer_depth: float = config_model["mixed_layer_depth"]["default"],
        coastline_action: str = config_model["coastline_action"]["default"],
        seafloor_action: str = config_model["seafloor_action"]["default"],
        max_speed: int = config_model["max_speed"]["default"],
        wind_drift_factor: float = config_model["wind_drift_factor"]["default"],
        wind_drift_depth: float = config_model["wind_drift_depth"]["default"],
        vertical_mixing_timestep: float = config_model["vertical_mixing_timestep"][
            "default"
        ],
        object_type: str = config_model["object_type"]["default"],
        diameter: float = config_model["diameter"]["default"],
        neutral_buoyancy_salinity: float = config_model["neutral_buoyancy_salinity"][
            "default"
        ],
        stage_fraction: float = config_model["stage_fraction"]["default"],
        hatched: float = config_model["hatched"]["default"],
        length: float = config_model["length"]["default"],
        weight: float = config_model["weight"]["default"],
        oil_type: str = config_model["oil_type"]["default"],
        m3_per_hour: float = config_model["m3_per_hour"]["default"],
        oil_film_thickness: float = config_model["oil_film_thickness"]["default"],
        droplet_size_distribution: str = config_model["droplet_size_distribution"][
            "default"
        ],
        droplet_diameter_mu: float = config_model["droplet_diameter_mu"]["default"],
        droplet_diameter_sigma: float = config_model["droplet_diameter_sigma"][
            "default"
        ],
        droplet_diameter_min_subsea: float = config_model[
            "droplet_diameter_min_subsea"
        ]["default"],
        droplet_diameter_max_subsea: float = config_model[
            "droplet_diameter_max_subsea"
        ]["default"],
        emulsification: bool = config_model["emulsification"]["default"],
        dispersion: bool = config_model["dispersion"]["default"],
        evaporation: bool = config_model["evaporation"]["default"],
        update_oilfilm_thickness: bool = config_model["update_oilfilm_thickness"][
            "default"
        ],
        biodegradation: bool = config_model["biodegradation"]["default"],
        log: str = config_model["log"]["default"],
        **kw,
    ) -> None:
        """Inputs for OpenDrift model."""

        # get all named parameters input to ParticleTrackingManager class
        from inspect import signature

        sig = signature(OpenDriftModel)

        # initialize all class attributes to None without triggering the __setattr__ method
        # which does a bunch more stuff
        for key in sig.parameters.keys():
            self.__dict__[key] = None

        self.__dict__["config_model"] = config_model

        model = "opendrift"

        if log == "low":
            self.__dict__["loglevel"] = 20
        elif log == "high":
            self.__dict__["loglevel"] = 0

        # need drift_model defined for the log to work properly for both manager and model
        # so do this before super initialization
        self.__dict__["drift_model"] = drift_model

        # do this right away so I can query the object
        if self.drift_model == "Leeway":
            o = Leeway(loglevel=self.loglevel)

        elif self.drift_model == "OceanDrift":
            o = OceanDrift(loglevel=self.loglevel)

        elif self.drift_model == "LarvalFish":
            o = LarvalFish(loglevel=self.loglevel)

        elif self.drift_model == "OpenOil":
            o = OpenOil(loglevel=self.loglevel, weathering_model="noaa")

        else:
            raise ValueError(f"Drifter model {self.drift_model} is not recognized.")

        self.__dict__["o"] = o

        self.__dict__["logger"] = logging.getLogger(
            model
        )  # use this syntax to avoid __setattr__

        super().__init__(model, **kw)

        # Extra keyword parameters are not currently allowed so they might be a typo
        if len(self.kw) > 0:
            raise KeyError(f"Unknown input parameter(s) {self.kw} input.")

        # Note that you can see configuration possibilities for a given model with
        # o.list_configspec()
        # You can check the metadata for a given configuration with (min/max/default/type)
        # o.get_configspec('vertical_mixing:timestep')
        # You can check required variables for a model with
        # o.required_variables

        # Set all attributes which will trigger some checks and changes in __setattr__
        # these will also update "value" in the config dict
        for key in sig.parameters.keys():
            # no need to run through for init if value is None (already set to None)
            if locals()[key] is not None:
                self.__setattr__(key, locals()[key])

    def calc_known_horizontal_diffusivity(self):
        """Calculate horizontal diffusivity based on known ocean_model."""

        # dx: approximate horizontal grid resolution (meters), used to calculate horizontal diffusivity
        if self.ocean_model == "NWGOA":
            dx = 1500
        elif "CIOFS" in self.ocean_model:
            dx = 100

        # horizontal diffusivity is calculated based on the mean horizontal grid resolution
        # for the model being used.
        # 0.1 is a guess for the magnitude of velocity being missed in the models, the sub-gridscale velocity
        sub_gridscale_velocity = 0.1
        horizontal_diffusivity = sub_gridscale_velocity * dx
        return horizontal_diffusivity

    def __setattr_model__(self, name: str, value) -> None:
        """Implement my own __setattr__ but here to enforce actions."""

        # don't allow drift_model to be reset, have to reinitialize object instead
        # check for type of m.o and drift_model matching to enforce this
        if (name in ["o", "drift_model"]) and (
            self.drift_model not in str(type(self.o))
        ):
            raise KeyError(
                "Can't overwrite `drift_model`; instead initialize OpenDriftModel with desired drift_model."
            )

        # create/update "value" keyword in config to keep it up to date
        if (
            self.config_model is not None
            # and name != "config_model"
            # and name != "config_ptm"
            and name in self.config_model.keys()
        ):
            self.config_model[name]["value"] = value
        self._update_config()

        if name in ["ocean_model", "horizontal_diffusivity"]:

            # just set the value and move on if purposely setting a non-None value
            # of horizontal_diffusivity; specifying this for clarity (already set
            # value above).
            if name == "horizontal_diffusivity" and value is not None:
                self.logger.info(
                    f"Setting horizontal_diffusivity to user-selected value {value}."
                )

            # in all other cases that ocean_model is a known model, want to use the
            # grid-dependent value
            elif self.ocean_model in _KNOWN_MODELS:

                hdiff = self.calc_known_horizontal_diffusivity()
                self.logger.info(
                    f"Setting horizontal_diffusivity parameter to one tuned to reader model of value {hdiff}."
                )
                # when editing the __dict__ directly have to also update config_model
                self.__dict__["horizontal_diffusivity"] = hdiff
                self.config_model["horizontal_diffusivity"]["value"] = hdiff

            # if user not using a known ocean_model, change horizontal_diffusivity from None to 0
            # so it has a value. User can subsequently overwrite it too.
            elif (
                self.ocean_model not in _KNOWN_MODELS
                and self.horizontal_diffusivity is None
            ):

                self.logger.info(
                    """Since ocean_model is user-input, changing horizontal_diffusivity parameter from None to 0.0.
                    You can also set it to a specific value with `m.horizontal_diffusivity=[number]`."""
                )

                self.__dict__["horizontal_diffusivity"] = 0
                self.config_model["horizontal_diffusivity"]["value"] = 0

        # turn on other things if using stokes_drift
        if name == "stokes_drift" and value:
            if self.drift_model != "Leeway":
                self.o.set_config("drift:use_tabularised_stokes_drift", True)
            # self.o.set_config('drift:tabularised_stokes_drift_fetch', '25000')  # default
            # self.o.set_config('drift:stokes_drift_profile', 'Phillips')  # default

        # too soon to do this, need to run it later
        # Leeway model doesn't have this option built in
        if name in ["surface_only", "drift_model"]:
            if self.surface_only and self.drift_model != "Leeway":
                self.logger.info("Truncating model output below 0.5 m.")
                self.o.set_config("drift:truncate_ocean_model_below_m", 0.5)
            elif (
                not self.surface_only
                and self.drift_model != "Leeway"
                and self.show_config(key="drift:truncate_ocean_model_below_m")["value"]
                is not None
            ):
                self.logger.info("Un-truncating model output below 0.5 m.")
                self.o.set_config("drift:truncate_ocean_model_below_m", None)

        # Leeway doesn't have this option available
        if name == "do3D" and not value and self.drift_model != "Leeway":
            self.logger.info("do3D is False so disabling vertical motion.")
            self.o.disable_vertical_motion()
        elif name == "do3D" and not value and self.drift_model == "Leeway":
            self.logger.info(
                "do3D is False but drift_model is Leeway so doing nothing."
            )

        if name == "do3D" and value and self.drift_model != "Leeway":
            self.logger.info("do3D is True so turning on vertical advection.")
            self.o.set_config("drift:vertical_advection", True)
        elif name == "do3D" and value and self.drift_model == "Leeway":
            self.logger.info(
                "do3D is True but drift_model is Leeway so changing do3D to False."
            )
            self.do3D = False

        # if drift_model is LarvalFish, vertical_mixing has to be True
        if name == "vertical_mixing" and not value and self.drift_model == "LarvalFish":
            raise ValueError(
                "drift_model is LarvalFish which always has vertical mixing on in OpenDrift so vertical_mixing must be True."
            )

        # if drift_model is LarvalFish, surface_only can't be True
        if name == "surface_only" and value and self.drift_model == "LarvalFish":
            raise ValueError(
                "drift_model is LarvalFish which is always 3D in OpenDrift so surface_only must be False."
            )

        # if drift_model is LarvalFish, do3D has to be True
        if name == "do3D" and not value and self.drift_model == "LarvalFish":
            raise ValueError(
                "drift_model is LarvalFish which is always 3D in OpenDrift so do3D must be True."
            )

        # Make sure vertical_mixing_timestep equals None if vertical_mixing False
        if name in ["vertical_mixing", "vertical_mixing_timestep"]:
            if not self.vertical_mixing:
                self.logger.info(
                    "vertical_mixing is False, so setting value of vertical_mixing_timestep "
                    "to None."
                )
                self.__dict__["vertical_mixing_timestep"] = None
                self.config_model["vertical_mixing_timestep"]["value"] = None

        # Make sure diffusivitymodel equals default value if vertical_mixing False
        if name in ["vertical_mixing", "diffusivitymodel"]:
            dmodeldef = self.config_model["diffusivitymodel"]["default"]
            if (
                not self.vertical_mixing
                and self.diffusivitymodel != dmodeldef
                and self.diffusivitymodel is not None
            ):
                self.logger.info(
                    "vertical_mixing is False, so resetting value of diffusivitymodel to default and not using."
                )
                self.__dict__["diffusivitymodel"] = dmodeldef
                self.config_model["diffusivitymodel"]["value"] = dmodeldef

        # Make sure mixed_layer_depth equals default value if vertical_mixing False
        if name in ["vertical_mixing", "mixed_layer_depth"]:
            mlddef = self.config_model["mixed_layer_depth"]["default"]
            if (
                not self.vertical_mixing
                and self.mixed_layer_depth != mlddef
                and self.mixed_layer_depth is not None
            ):
                self.logger.info(
                    "vertical_mixing is False, so resetting value of mixed_layer_depth to default and not using."
                )
                self.__dict__["mixed_layer_depth"] = mlddef
                self.config_model["mixed_layer_depth"]["value"] = mlddef

        # make sure user isn't try to use Leeway or LarvalFish and "wind_drift_factor" at the same time
        if name == "wind_drift_factor":
            if self.drift_model in ["Leeway", "LarvalFish"]:
                self.logger.info(
                    "wind_drift_factor cannot be used with Leeway or LarvalFish models, "
                    "so setting to None."
                )
                self.__dict__["wind_drift_factor"] = None
                self.config_model["wind_drift_factor"]["value"] = None

        # make sure user isn't try to use Leeway or LarvalFish models and "wind_drift_depth" at the same time
        if name == "wind_drift_depth":
            if self.drift_model in ["Leeway", "LarvalFish"]:
                self.logger.info(
                    "wind_drift_depth cannot be used with Leeway or LarvalFish models, "
                    "so setting to None."
                )
                self.__dict__["wind_drift_depth"] = None
                self.config_model["wind_drift_depth"]["value"] = None

        # make sure user isn't try to use Leeway and "stokes_drift" at the same time
        if name == "stokes_drift":
            if self.drift_model == "Leeway" and self.stokes_drift:
                self.logger.info(
                    "stokes_drift cannot be used with Leeway model, so changing to False."
                )
                self.__dict__["stokes_drift"] = False
                self.config_model["stokes_drift"]["value"] = False

        # Add export variables for certain drift_model values
        # drift_model is always set initially only
        if name == "export_variables" and self.drift_model == "OpenOil":
            oil_vars = [
                "mass_oil",
                "density",
                "mass_evaporated",
                "mass_dispersed",
                "mass_biodegraded",
                "viscosity",
                "water_fraction",
            ]
            self.__dict__["export_variables"] += oil_vars
            self.config_model["export_variables"]["value"] += oil_vars
        elif name == "export_variables" and self.drift_model == "Leeway":
            vars = ["object_type"]
            self.__dict__["export_variables"] += vars
            self.config_model["export_variables"]["value"] += vars
        elif name == "export_variables" and self.drift_model == "LarvalFish":
            vars = [
                "diameter",
                "neutral_buoyancy_salinity",
                "stage_fraction",
                "hatched",
                "length",
                "weight",
            ]
            self.__dict__["export_variables"] += vars
            self.config_model["export_variables"]["value"] += vars

        self._update_config()

    def run_add_reader(
        self,
        ds=None,
        name=None,
        oceanmodel_lon0_360=False,
        standard_name_mapping=None,
    ):
        """Might need to cache this if its still slow locally.

        Parameters
        ----------
        ds : xr.Dataset, optional
            Previously-opened Dataset containing ocean model output, if user wants to input
            unknown reader information.
        name : str, optional
            If ds is input, user can also input name of ocean model, otherwise will be called "user_input".
        oceanmodel_lon0_360 : bool
            True if ocean model longitudes span 0 to 360 instead of -180 to 180.
        standard_name_mapping : dict
            Mapping of model variable names to standard names.
        """

        if (
            self.ocean_model not in _KNOWN_MODELS
            and self.ocean_model != "test"
            and ds is None
        ):
            raise ValueError(
                "ocean_model must be a known model or user must input a Dataset."
            )

        standard_name_mapping = standard_name_mapping or {}

        if ds is not None:
            if name is None:
                self.ocean_model = "user_input"
            else:
                self.ocean_model = name

        if self.ocean_model == "test":
            pass
            # oceanmodel_lon0_360 = True
            # loc = "test"
            # kwargs_xarray = dict()

        elif self.ocean_model is not None or ds is not None:
            if self.ocean_model_local:
                self.logger.info(
                    f"Using local output for ocean_model {self.ocean_model}"
                )
            else:
                self.logger.info(
                    f"Using remote output for ocean_model {self.ocean_model}"
                )

            # set drop_vars initial values based on the PTM settings, then add to them for the specific model
            drop_vars = []
            # don't need w if not 3D movement
            if not self.do3D:
                drop_vars += ["w"]
                self.logger.info("Dropping vertical velocity (w) because do3D is False")
            else:
                self.logger.info("Retaining vertical velocity (w) because do3D is True")

            # don't need winds if stokes drift, wind drift, added wind_uncertainty, and vertical_mixing are off
            # It's possible that winds aren't required for every OpenOil simulation but seems like
            # they would usually be required and the cases are tricky to navigate so also skipping for that case.
            if (
                not self.stokes_drift
                and self.wind_drift_factor == 0
                and self.wind_uncertainty == 0
                and self.drift_model != "OpenOil"
                and not self.vertical_mixing
            ):
                drop_vars += ["Uwind", "Vwind", "Uwind_eastward", "Vwind_northward"]
                self.logger.info(
                    "Dropping wind variables because stokes_drift, wind_drift_factor, wind_uncertainty, and vertical_mixing are all off and drift_model is not 'OpenOil'"
                )
            else:
                self.logger.info(
                    "Retaining wind variables because stokes_drift, wind_drift_factor, wind_uncertainty, or vertical_mixing are on or drift_model is 'OpenOil'"
                )

            # only keep salt and temp for LarvalFish or OpenOil
            if self.drift_model not in ["LarvalFish", "OpenOil"]:
                drop_vars += ["salt", "temp"]
                self.logger.info(
                    "Dropping salt and temp variables because drift_model is not LarvalFish nor OpenOil"
                )
            else:
                self.logger.info(
                    "Retaining salt and temp variables because drift_model is LarvalFish or OpenOil"
                )

            # keep some ice variables for OpenOil (though later see if these are used)
            if self.drift_model != "OpenOil":
                drop_vars += ["aice", "uice_eastward", "vice_northward"]
                self.logger.info(
                    "Dropping ice variables because drift_model is not OpenOil"
                )
            else:
                self.logger.info(
                    "Retaining ice variables because drift_model is OpenOil"
                )

            # if using static masks, drop wetdry masks.
            # if using wetdry masks, drop static masks.
            if self.use_static_masks:
                standard_name_mapping.update({"mask_rho": "land_binary_mask"})
                drop_vars += ["wetdry_mask_rho", "wetdry_mask_u", "wetdry_mask_v"]
                self.logger.info(
                    "Dropping wetdry masks because using static masks instead."
                )
            else:
                standard_name_mapping.update({"wetdry_mask_rho": "land_binary_mask"})
                drop_vars += ["mask_rho", "mask_u", "mask_v", "mask_psi"]
                self.logger.info(
                    "Dropping mask_rho, mask_u, mask_v, mask_psi because using wetdry masks instead."
                )

            if self.ocean_model == "NWGOA":
                oceanmodel_lon0_360 = True

                standard_name_mapping.update(
                    {
                        "u_eastward": "x_sea_water_velocity",
                        "v_northward": "y_sea_water_velocity",
                        # NWGOA, there are east/north oriented and will not be rotated
                        # because "east" "north" in variable names
                        "Uwind_eastward": "x_wind",
                        "Vwind_northward": "y_wind",
                    }
                )

                # remove all other grid masks because variables are all on rho grid
                drop_vars += [
                    "hice",
                    "hraw",
                    "snow_thick",
                ]

                loc_local = "/mnt/depot/data/packrat/prod/aoos/nwgoa/processed/nwgoa_kerchunk.parq"
                loc_remote = (
                    "http://xpublish-nwgoa.srv.axds.co/datasets/nwgoa_all/zarr/"
                )

            elif "CIOFS" in self.ocean_model:
                oceanmodel_lon0_360 = False

                drop_vars += [
                    "wetdry_mask_psi",
                ]
                if self.ocean_model == "CIOFS":

                    loc_local = "/mnt/vault/ciofs/HINDCAST/ciofs_kerchunk.parq"
                    loc_remote = "http://xpublish-ciofs.srv.axds.co/datasets/ciofs_hindcast/zarr/"

                elif self.ocean_model == "CIOFSOP":

                    standard_name_mapping.update(
                        {
                            "u_eastward": "x_sea_water_velocity",
                            "v_northward": "y_sea_water_velocity",
                        }
                    )

                    loc_local = "/mnt/depot/data/packrat/prod/noaa/coops/ofs/aws_ciofs/processed/aws_ciofs_kerchunk.parq"
                    loc_remote = "https://thredds.aoos.org/thredds/dodsC/AWS_CIOFS.nc"

            elif self.ocean_model == "user_input":

                # check for case that self.use_static_masks False (which is the default)
                # but user input doesn't have wetdry masks
                # then raise exception and tell user to set use_static_masks True
                if "wetdry_mask_rho" not in ds.data_vars and not self.use_static_masks:
                    raise ValueError(
                        "User input does not have wetdry_mask_rho variable. Set use_static_masks True to use static masks instead."
                    )

                ds = ds.drop_vars(drop_vars, errors="ignore")

            # if local and not a user-input ds
            if ds is None:
                if self.ocean_model_local:

                    ds = xr.open_dataset(
                        loc_local,
                        engine="kerchunk",
                        chunks={},
                        drop_variables=drop_vars,
                        decode_times=False,
                    )

                # otherwise remote
                else:
                    if ".nc" in loc_remote:
                        ds = xr.open_dataset(
                            loc_remote,
                            chunks={},
                            drop_variables=drop_vars,
                            decode_times=False,
                        )
                    else:
                        ds = xr.open_zarr(
                            loc_remote,
                            chunks={},
                            drop_variables=drop_vars,
                            decode_times=False,
                        )

            # For NWGOA, need to calculate wetdry mask from a variable
            if self.ocean_model == "NWGOA" and not self.use_static_masks:
                ds["wetdry_mask_rho"] = (~ds.zeta.isnull()).astype(int)

            # For CIOFSOP need to rename u/v to have "East" and "North" in the variable names
            # so they aren't rotated in the ROMS reader (the standard names have to be x/y not east/north)
            elif self.ocean_model == "CIOFSOP":
                ds = ds.rename_vars({"urot": "u_eastward", "vrot": "v_northward"})
                # grid = xr.open_dataset("/mnt/vault/ciofs/HINDCAST/nos.ciofs.romsgrid.nc")
                # ds["angle"] = grid["angle"]

            try:
                units = ds.ocean_time.attrs["units"]
            except KeyError:
                units = ds.ocean_time.encoding["units"]
            datestr = units.split("since ")[1]
            units_date = pd.Timestamp(datestr)

            # use reader start time if not otherwise input
            if self.start_time is None:
                self.logger.info("setting reader start_time as simulation start_time")
                # self.start_time = reader.start_time
                # convert using pandas instead of netCDF4
                self.start_time = units_date + pd.to_timedelta(
                    ds.ocean_time[0].values, unit="s"
                )
            # narrow model output to simulation time if possible before sending to Reader
            if self.start_time is not None and self.end_time is not None:
                dt_model = float(
                    ds.ocean_time[1] - ds.ocean_time[0]
                )  # time step of the model output in seconds
                start_time_num = (self.start_time - units_date).total_seconds()
                # want to include the next ocean model output after the last drifter simulation time
                end_time_num = (self.end_time - units_date).total_seconds() + dt_model
                ds = ds.sel(ocean_time=slice(start_time_num, end_time_num))
                self.logger.info("Narrowed model output to simulation time")
            else:
                raise ValueError(
                    "start_time and end_time must be set to narrow model output to simulation time"
                )

            # save interpolators to save time
            cache_dir = Path(
                appdirs.user_cache_dir(
                    appname="particle-tracking-manager", appauthor="axiom-data-science"
                )
            )
            cache_dir.mkdir(parents=True, exist_ok=True)
            self.cache_dir = cache_dir
            self.interpolator_filename = cache_dir / Path(
                f"{self.ocean_model}_interpolator"
            )
            reader = reader_ROMS_native.Reader(
                filename=ds,
                name=self.ocean_model,
                standard_name_mapping=standard_name_mapping,
                save_interpolator=True,
                interpolator_filename=self.interpolator_filename,
            )

            self.o.add_reader([reader])
            self.reader = reader
            # can find reader at manager.o.env.readers[self.ocean_model]

            self.oceanmodel_lon0_360 = oceanmodel_lon0_360

        else:
            raise ValueError("reader did not set an ocean_model")

    @property
    def seed_kws(self):
        """Gather seed input kwargs.

        This could be run more than once.
        """

        already_there = [
            "seed:number",
            "seed:z",
            "seed:seafloor",
            "seed:droplet_diameter_mu",
            "seed:droplet_diameter_min_subsea",
            "seed:droplet_size_distribution",
            "seed:droplet_diameter_sigma",
            "seed:droplet_diameter_max_subsea",
            "seed:object_type",
            "seed_flag",
            "drift:use_tabularised_stokes_drift",
            "drift:vertical_advection",
            "drift:truncate_ocean_model_below_m",
        ]

        _seed_kws = {
            "time": self.start_time.to_pydatetime()
            if self.start_time is not None
            else None,
            "z": self.z,
        }

        # update seed_kws with drift_model-specific seed parameters
        seedlist = self.drift_model_config(prefix="seed")
        seedlist = [(one, two) for one, two in seedlist if one not in already_there]
        seedlist = [(one.replace("seed:", ""), two) for one, two in seedlist]
        _seed_kws.update(seedlist)

        if self.seed_flag == "elements":
            # add additional seed parameters
            _seed_kws.update(
                {
                    "lon": self.lon,
                    "lat": self.lat,
                    "radius": self.radius,
                    "radius_type": self.radius_type,
                }
            )

        elif self.seed_flag == "geojson":

            # geojson needs string representation of time
            _seed_kws["time"] = (
                self.start_time.isoformat() if self.start_time is not None else None
            )

        self._seed_kws = _seed_kws
        return self._seed_kws

    def run_seed(self):
        """Actually seed drifters for model."""

        if self.seed_flag == "elements":

            self.o.seed_elements(**self.seed_kws)

        elif self.seed_flag == "geojson":

            # geojson needs string representation of time
            self.seed_kws["time"] = self.start_time.isoformat()
            self.geojson["properties"] = self.seed_kws
            json_string_dumps = json.dumps(self.geojson)
            self.o.seed_from_geojson(json_string_dumps)

        else:
            raise ValueError(f"seed_flag {self.seed_flag} not recognized.")

        self.initial_drifters = self.o.elements_scheduled

    def run_drifters(self):
        """Run the drifters!"""

        if self.steps is None and self.duration is None and self.end_time is None:
            raise ValueError(
                "Exactly one of steps, duration, or end_time must be input and not be None."
            )

        if self.run_forward:
            timedir = 1
        else:
            timedir = -1

        # drop non-OpenDrift parameters now so they aren't brought into simulation (they mess up the write step)
        full_config = copy.deepcopy(self._config)  # save
        config_input_to_opendrift = {
            k: full_config[k] for k in self._config_orig.keys()
        }

        self.o._config = config_input_to_opendrift  # only OpenDrift config

        output_file_initial = (
            f"{self.output_file}_initial"
            or f"output-results_{datetime.datetime.utcnow():%Y-%m-%dT%H%M:%SZ}.nc"
        )

        self.o.run(
            time_step=timedir * self.time_step,
            time_step_output=self.time_step_output,
            steps=self.steps,
            export_variables=self.export_variables,
            outfile=output_file_initial,
        )

        self.o._config = full_config  # reinstate config

        # open outfile file and add config to it
        # config can't be present earlier because it breaks opendrift
        ds = xr.open_dataset(output_file_initial)
        for k, v in self.drift_model_config():
            if isinstance(v, (bool, type(None), pd.Timestamp, pd.Timedelta)):
                v = str(v)
            ds.attrs[f"ptm_config_{k}"] = v

        # Make new output file
        output_file = (
            self.output_file
            or f"output-results_{datetime.datetime.utcnow():%Y-%m-%dT%H%M:%SZ}.nc"
        )

        ds.to_netcdf(output_file)

        # update with new path name
        self.o.outfile_name = output_file

        try:
            # remove initial file to save space
            os.remove(output_file_initial)
        except PermissionError:
            # windows issue
            pass

    @property
    def _config(self):
        """Surface the model configuration."""

        # save for reinstatement when running the drifters
        if self._config_orig is None:
            self._config_orig = copy.deepcopy(self.o._config)

        return self.o._config

    def _add_ptm_config(self):
        """Add PTM config to overall config."""

        dict1 = copy.deepcopy(self._config)
        dict2 = copy.deepcopy(self.config_ptm)

        # dictB has the od_mapping version of the keys of dict2
        # e.g.  'processes:emulsification' instead of 'emulsification'
        # dictB values are the OpenDriftModel config parameters with config_od parameters added on top
        dictB = {
            v["od_mapping"]: (
                {**dict1[v["od_mapping"]], **v}
                if "od_mapping" in v and v["od_mapping"] in dict1.keys()
                else {**v}
            )
            for k, v in dict2.items()
            if "od_mapping" in v
        }

        # dictC is the same as dictB except the names are the PTM/OpenDriftModel names instead of the
        # original OpenDrift names
        dictC = {
            k: {**dict1[v["od_mapping"]], **v}
            if "od_mapping" in v and v["od_mapping"] in dict1.keys()
            else {**v}
            for k, v in dict2.items()
            if "od_mapping" in v
        }

        # this step copies in parameter info from config_ptm to _config
        self._config.update(dict2)

        # this step brings config overrides from config_ptm into the overall config
        self._config.update(dictB)
        # this step brings other model config (plus additions from mapped parameter config) into the overall config
        self._config.update(dictC)
        # # this step brings other model config into the overall config
        # self._config.update(dict2)

    def _add_model_config(self):
        """Goal is to combine the config both directions:

        * override OpenDrift config defaults with those from opendrift_config as well
          as include extra information like ptm_level
        * bring OpenDrift config parameter metadata into config_model so application
          could query it to get the ranges, options, etc.
        """

        dict1 = copy.deepcopy(self._config)
        dict2 = copy.deepcopy(self.config_model)

        # dictB has the od_mapping version of the keys of dict2
        # e.g.  'processes:emulsification' instead of 'emulsification'
        # dictB values are the OpenDrift config parameters with config_od parameters added on top
        dictB = {
            v["od_mapping"]: {**dict1[v["od_mapping"]], **v}
            if "od_mapping" in v and v["od_mapping"] in dict1.keys()
            else {**v}
            for k, v in dict2.items()
            if "od_mapping" in v
        }

        # dictC is the same as dictB except the names are the PTM/OpenDriftModel names instead of the
        # original OpenDrift names
        dictC = {
            k: {**dict1[v["od_mapping"]], **v}
            if "od_mapping" in v and v["od_mapping"] in dict1.keys()
            else {**v}
            for k, v in dict2.items()
            if "od_mapping" in v
        }

        # this step copies in parameter info from config_ptm to _config
        self._config.update(dict2)

        # this step brings config overrides from config_model into the overall config
        self._config.update(dictB)
        # this step brings other model config (plus additions from mapped parameter config) into the overall config
        self._config.update(dictC)

    def all_export_variables(self):
        """Output list of all possible export variables."""

        vars = (
            list(self.o.elements.variables.keys())
            + ["trajectory", "time"]
            + list(self.o.required_variables.keys())
        )

        return vars

    def export_variables(self):
        """Output list of all actual export variables."""

        return self.o.export_variables

    def drift_model_config(self, ptm_level=[1, 2, 3], prefix=""):
        """Show config for this drift model selection.

        This shows all PTM-controlled parameters for the OpenDrift
        drift model selected and their current values, at the selected ptm_level
        of importance. It includes some additional configuration parameters
        that are indirectly controlled by PTM parameters.

        Parameters
        ----------
        ptm_level : int, list, optional
            Options are 1, 2, 3, or lists of combinations. Use [1,2,3] for all.
            Default is 1.
        prefix : str, optional
            prefix to search config for, only for OpenDrift parameters (not PTM).
        """

        outlist = [
            (key, value_dict["value"])
            for key, value_dict in self.show_config(
                substring=":", ptm_level=ptm_level, level=[1, 2, 3], prefix=prefix
            ).items()
            if "value" in value_dict and value_dict["value"] is not None
        ]

        # also PTM config parameters that are separate from OpenDrift parameters
        outlist2 = [
            (key, value_dict["value"])
            for key, value_dict in self.show_config(
                ptm_level=ptm_level, prefix=prefix
            ).items()
            if "od_mapping" not in value_dict
            and "value" in value_dict
            and value_dict["value"] is not None
        ]

        # extra parameters that are not in the config_model but are set by PTM indirectly
        extra_keys = [
            "drift:vertical_advection",
            "drift:truncate_ocean_model_below_m",
            "drift:use_tabularised_stokes_drift",
        ]
        outlist += [
            (key, self.show_config(key=key)["value"])
            for key in extra_keys
            if "value" in self.show_config(key=key)
        ]

        return outlist + outlist2

    def get_configspec(self, prefix, substring, excludestring, level, ptm_level):
        """Copied from OpenDrift, then modified."""

        if not isinstance(level, list) and level is not None:
            level = [level]
        if not isinstance(ptm_level, list) and ptm_level is not None:
            ptm_level = [ptm_level]

        # check for prefix or substring comparison
        configspec = {
            k: v
            for (k, v) in self._config.items()
            if k.startswith(prefix) and substring in k and excludestring not in k
        }

        if level is not None:
            # check for levels (if present)
            configspec = {
                k: v
                for (k, v) in configspec.items()
                if "level" in configspec[k] and configspec[k]["level"] in level
            }

        if ptm_level is not None:
            # check for ptm_levels (if present)
            configspec = {
                k: v
                for (k, v) in configspec.items()
                if "ptm_level" in configspec[k]
                and configspec[k]["ptm_level"] in ptm_level
            }

        return configspec

    def show_config_model(
        self,
        key=None,
        prefix="",
        level=None,
        ptm_level=None,
        substring="",
        excludestring="excludestring",
    ) -> dict:
        """Show configuring for the drift model selected in configuration.

        Runs configuration for you if it hasn't yet been run.

        Parameters
        ----------
        key : str, optional
            If input, show configuration for just that key.
        prefix : str, optional
            prefix to search config for, only for OpenDrift parameters (not PTM).
        level : int, list, optional
            Limit search by level:

            * CONFIG_LEVEL_ESSENTIAL = 1
            * CONFIG_LEVEL_BASIC = 2
            * CONFIG_LEVEL_ADVANCED = 3

            e.g. 1, [1,2], [1,2,3]
        ptm_level : int, list, optional
            Limit search by level:

            * Surface to user = 1
            * Medium surface to user = 2
            * Surface but bury = 3

            e.g. 1, [1,2], [1,2,3]. To access all PTM parameters search for
            `ptm_level=[1,2,3]`.
        substring : str, optional
            If input, show configuration that contains that substring.
        excludestring : str, optional
            configuration parameters are not shown if they contain this string.

        Examples
        --------
        Show all possible configuration for the previously-selected drift model:

        >>> manager.show_config()

        Show configuration with a specific prefix:

        >>> manager.show_config(prefix="seed")

        Show configuration matching a substring:

        >>> manager.show_config(substring="stokes")

        Show configuration at a specific level (from OpenDrift):

        >>> manager.show_config(level=1)

        Show all OpenDrift configuration:

        >>> manager.show_config(level=[1,2,3])

        Show configuration for only PTM-specified parameters:

        >>> manager.show_config(ptm_level=[1,2,3])

        Show configuration for a specific PTM level:

        >>> manager.show_config(ptm_level=2)

        Show configuration for a single key:

        >>> manager.show_config("seed:oil_type")

        Show configuration for parameters that are both OpenDrift and PTM-modified:

        >>> m.show_config(ptm_level=[1,2,3], level=[1,2,3])

        """

        if key is not None:
            prefix = key

        output = self.get_configspec(
            prefix=prefix,
            level=level,
            ptm_level=ptm_level,
            substring=substring,
            excludestring=excludestring,
        )

        if key is not None:
            if key in output:
                return output[key]
            else:
                return output
        else:
            return output

    def reader_metadata(self, key):
        """allow manager to query reader metadata."""

        if not self.has_added_reader:
            raise ValueError("reader has not been added yet.")
        return self.o.env.readers[self.ocean_model].__dict__[key]

    @property
    def outfile_name(self):
        """Output file name."""

        return self.o.outfile_name



"""From Copilot"""

import unittest

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from particle_tracking_manager.models.opendrift.opendrift import (
    OpenDriftModel,
    config_model,
)


class TestOpenDriftModel(unittest.TestCase):
    def setUp(self):
        self.odm = OpenDriftModel()

    def test_init(self):
        self.assertEqual(self.odm.drift_model, config_model["drift_model"]["default"])
        self.assertEqual(self.odm.radius, config_model["radius"]["default"])
        self.assertEqual(self.odm.radius_type, config_model["radius_type"]["default"])
        # self.assertEqual(self.odm.horizontal_diffusivity, config_model["horizontal_diffusivity"]["default"])
        self.assertEqual(
            self.odm.use_auto_landmask, config_model["use_auto_landmask"]["default"]
        )
        self.assertEqual(self.odm.diffusivitymodel, "windspeed_Large1994")
        self.assertEqual(self.odm.stokes_drift, config_model["stokes_drift"]["default"])
        self.assertEqual(
            self.odm.mixed_layer_depth, config_model["mixed_layer_depth"]["default"]
        )
        self.assertEqual(
            self.odm.coastline_action, config_model["coastline_action"]["default"]
        )
        self.assertEqual(self.odm.max_speed, config_model["max_speed"]["default"])
        self.assertEqual(
            self.odm.wind_drift_factor, config_model["wind_drift_factor"]["default"]
        )
        self.assertEqual(
            self.odm.wind_drift_depth, config_model["wind_drift_depth"]["default"]
        )
        self.assertEqual(
            self.odm.vertical_mixing_timestep,
            None,
        )
        self.assertEqual(self.odm.object_type, config_model["object_type"]["default"])
        self.assertEqual(self.odm.diameter, config_model["diameter"]["default"])
        self.assertEqual(
            self.odm.neutral_buoyancy_salinity,
            config_model["neutral_buoyancy_salinity"]["default"],
        )
        self.assertEqual(
            self.odm.stage_fraction, config_model["stage_fraction"]["default"]
        )
        self.assertEqual(self.odm.hatched, config_model["hatched"]["default"])
        self.assertEqual(self.odm.length, config_model["length"]["default"])
        self.assertEqual(self.odm.weight, config_model["weight"]["default"])
        self.assertEqual(self.odm.oil_type, config_model["oil_type"]["default"])
        self.assertEqual(self.odm.m3_per_hour, config_model["m3_per_hour"]["default"])
        self.assertEqual(
            self.odm.oil_film_thickness, config_model["oil_film_thickness"]["default"]
        )
        self.assertEqual(
            self.odm.droplet_size_distribution,
            config_model["droplet_size_distribution"]["default"],
        )
        self.assertEqual(
            self.odm.droplet_diameter_mu, config_model["droplet_diameter_mu"]["default"]
        )
        self.assertEqual(
            self.odm.droplet_diameter_sigma,
            config_model["droplet_diameter_sigma"]["default"],
        )
        self.assertEqual(
            self.odm.droplet_diameter_min_subsea,
            config_model["droplet_diameter_min_subsea"]["default"],
        )
        self.assertEqual(
            self.odm.droplet_diameter_max_subsea,
            config_model["droplet_diameter_max_subsea"]["default"],
        )
        self.assertEqual(
            self.odm.emulsification, config_model["emulsification"]["default"]
        )
        self.assertEqual(self.odm.dispersion, config_model["dispersion"]["default"])
        self.assertEqual(self.odm.evaporation, config_model["evaporation"]["default"])
        self.assertEqual(
            self.odm.update_oilfilm_thickness,
            config_model["update_oilfilm_thickness"]["default"],
        )
        self.assertEqual(
            self.odm.biodegradation, config_model["biodegradation"]["default"]
        )


ds = xr.Dataset(
    data_vars={
        "u": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "v": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "w": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "salt": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "temp": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "wetdry_mask_rho": (("ocean_time", "Z", "Y", "X"), np.zeros((2, 3, 2, 3))),
        "mask_rho": (("Y", "X"), np.zeros((2, 3))),
        "Uwind": (("ocean_time", "Y", "X"), np.zeros((2, 2, 3))),
        "Vwind": (("ocean_time", "Y", "X"), np.zeros((2, 2, 3))),
        "Cs_r": (("Z"), np.linspace(-1, 0, 3)),
        "hc": 16,
    },
    coords={
        "ocean_time": ("ocean_time", [0, 1], {"units": "seconds since 1970-01-01"}),
        "s_rho": (("Z"), np.linspace(-1, 0, 3)),
        "lon_rho": (("Y", "X"), np.array([[1, 2, 3], [1, 2, 3]])),
        "lat_rho": (("Y", "X"), np.array([[1, 1, 1], [2, 2, 2]])),
    },
)


class TestOpenDriftModel_OceanDrift_static_mask(unittest.TestCase):
    def setUp(self):
        self.model = OpenDriftModel(drift_model="OceanDrift", use_static_masks=True)

    def test_ocean_model_not_known_ds_None(self):
        self.model.ocean_model = "wrong_name"
        self.model.ds = None  # this is the default
        # need to input steps, duration, or end_time but don't here
        with pytest.raises(ValueError):
            self.model.add_reader(ds=ds)

    def test_drop_vars_do3D_true(self):
        self.model.do3D = True
        self.model.steps = 4
        self.model.add_reader(ds=ds)
        assert self.model.reader.variables == [
            "x_sea_water_velocity",
            "y_sea_water_velocity",
            "upward_sea_water_velocity",
            "land_binary_mask",
            "x_wind",
            "y_wind",
            "wind_speed",
            "sea_water_speed",
        ]

    def test_drop_vars_use_static_masks(self):
        self.model.do3D = False
        self.model.duration = pd.Timedelta("24h")
        self.model.add_reader(ds=ds)
        assert self.model.reader.variables == [
            "x_sea_water_velocity",
            "y_sea_water_velocity",
            "land_binary_mask",
            "x_wind",
            "y_wind",
            "wind_speed",
            "sea_water_speed",
        ]
        assert "mask_rho" in self.model.reader.Dataset.data_vars
        assert "wetdry_mask_rho" not in self.model.reader.Dataset.data_vars

    def test_drop_vars_no_wind(self):
        self.model.stokes_drift = False
        self.model.wind_drift_factor = 0
        self.model.wind_uncertainty = 0
        self.model.vertical_mixing = False
        self.model.end_time = pd.Timestamp("1970-01-01T02:00")
        self.model.add_reader(ds=ds)
        assert self.model.reader.variables == [
            "x_sea_water_velocity",
            "y_sea_water_velocity",
            "land_binary_mask",
            "sea_water_speed",
        ]


class TestOpenDriftModel_OceanDrift_wetdry_mask(unittest.TestCase):
    def setUp(self):
        self.model = OpenDriftModel(drift_model="OceanDrift", use_static_masks=False)

    def test_error_no_end_of_simulation(self):
        self.model.do3D = False
        # need to input steps, duration, or end_time but don't here
        with pytest.raises(ValueError):
            self.model.add_reader(ds=ds)

    def test_drop_vars_do3D_false(self):
        self.model.do3D = False
        self.model.steps = 4
        self.model.add_reader(ds=ds)
        assert self.model.reader.variables == [
            "x_sea_water_velocity",
            "y_sea_water_velocity",
            "land_binary_mask",
            "x_wind",
            "y_wind",
            "wind_speed",
            "sea_water_speed",
        ]
        assert "wetdry_mask_rho" in self.model.reader.Dataset.data_vars
        assert "mask_rho" not in self.model.reader.Dataset.data_vars


class TestOpenDriftModel_LarvalFish(unittest.TestCase):
    def setUp(self):
        self.model = OpenDriftModel(drift_model="LarvalFish", do3D=True)

    def test_drop_vars_wind(self):
        self.model.duration = pd.Timedelta("1h")
        self.model.add_reader(ds=ds)
        assert self.model.reader.variables == [
            "x_sea_water_velocity",
            "y_sea_water_velocity",
            "upward_sea_water_velocity",
            "sea_water_salinity",
            "sea_water_temperature",
            "land_binary_mask",
            "x_wind",
            "y_wind",
            "wind_speed",
            "sea_water_speed",
        ]


def test_drift_model():
    with pytest.raises(ValueError):
        m = OpenDriftModel(drift_model="not_a_real_model")


class TestTheManager(unittest.TestCase):
    def setUp(self):
        self.m = OpenDriftModel()
        # self.m.config_model = {"test_key": {"value": "old_value"}}
        # test m.drift_model_config()
        # also test in other sub tests

    def test_set_drift_model(self):
        """can't change the drift_model after class initialization"""
        with self.assertRaises(KeyError):
            self.m.drift_model = "new_model"

    def test_set_o(self):
        with self.assertRaises(KeyError):
            self.m.o = "new_o"

    def test_stokes_drift_true_not_leeway(self):
        self.m.stokes_drift = True
        assert self.m.show_config(key="drift:use_tabularised_stokes_drift")

    def test_surface_only_true_not_leeway(self):
        self.m.surface_only = True
        assert (
            self.m.show_config(key="drift:truncate_ocean_model_below_m")["value"] == 0.5
        )

    def test_surface_only_false_not_leeway(self):
        self.m.surface_only = False
        assert (
            self.m.show_config(key="drift:truncate_ocean_model_below_m")["value"]
            is None
        )

    def test_do3D_false_not_leeway(self):
        self.m.do3D = False
        assert not self.m.show_config(key="drift:vertical_advection")["value"]
        assert not self.m.show_config(key="drift:vertical_mixing")["value"]

    def test_do3D_true(self):
        self.m.do3D = True
        # assert self.m.show_config(key="drift:vertical_advection")["value"]

    def test_vertical_mixing_false_vertical_mixing_timestep_not_default(self):
        self.m.vertical_mixing = False
        self.m.vertical_mixing_timestep = 10
        d = self.m.show_config(key="vertical_mixing_timestep")
        assert d["value"] == None

    def test_vertical_mixing_false_diffusivitymodel_not_default(self):
        self.m.vertical_mixing = False
        self.m.diffusivitymodel = "not_default"
        d = self.m.show_config(key="diffusivitymodel")
        assert d["value"] == "windspeed_Large1994"

    def test_vertical_mixing_false_mixed_layer_depth_not_default(self):
        self.m.vertical_mixing = False
        self.m.mixed_layer_depth = 10
        d = self.m.show_config(key="mixed_layer_depth")
        assert d["value"] == 30


class TestOpenDriftModel_Leeway(unittest.TestCase):
    def setUp(self):
        self.m = OpenDriftModel(
            drift_model="Leeway", object_type=">PIW, scuba suit (face up)"
        )

    def test_leeway_model_wind_drift_factor_not_default(self):
        self.m.wind_drift_factor = 10
        d = self.m.show_config(key="wind_drift_factor")
        assert d["value"] == None
        assert self.m.object_type == ">PIW, scuba suit (face up)"
        assert self.m.o._config["object_type"]["value"] == ">PIW, scuba suit (face up)"

    def test_leeway_model_wind_drift_depth_not_default(self):
        self.m.wind_drift_depth = 10
        d = self.m.show_config(key="wind_drift_depth")
        assert d["value"] == None

    def test_leeway_model_stokes_drift_true(self):
        self.m.stokes_drift = True
        assert not self.m.stokes_drift
        # assert not self.m.show_config(key="stokes_drift")["value"]


def test_horizontal_diffusivity_logic():
    """Check logic for using default horizontal diff values for known models."""

    m = OpenDriftModel()
    m.ocean_model = "NWGOA"
    assert m.horizontal_diffusivity == 150.0  # known grid values
    m.ocean_model = "CIOFS"
    assert m.horizontal_diffusivity == 10.0  # known grid values

    # or can overwrite it in this order
    m.horizontal_diffusivity = 11
    assert m.horizontal_diffusivity == 11.0  # user-selected value

    m.ocean_model = "CIOFSOP"
    assert m.horizontal_diffusivity == 10.0  # known grid values

    m = OpenDriftModel(ocean_model="NWGOA")
    assert m.horizontal_diffusivity == 150.0  # known grid values


def test_LarvalFish_disallowed_settings():
    """LarvalFish is incompatible with some settings.

    LarvalFish has to always be 3D.
    """

    with pytest.raises(ValueError):
        m = OpenDriftModel(drift_model="LarvalFish", vertical_mixing=False)

    with pytest.raises(ValueError):
        m = OpenDriftModel(drift_model="LarvalFish", surface_only=True)

    m = OpenDriftModel(drift_model="LarvalFish", do3D=True)
    with pytest.raises(ValueError):
        m.surface_only = True

    with pytest.raises(ValueError):
        m = OpenDriftModel(drift_model="LarvalFish", do3D=False)


def test_LarvalFish_seeding():
    """Make sure special seed parameter comes through"""

    m = OpenDriftModel(
        drift_model="LarvalFish",
        lon=-151,
        lat=60,
        do3D=True,
        hatched=1,
        start_time="2021-01-01T00:00:00",
        use_auto_landmask=True,
    )
    m.seed()
    assert m.o.elements_scheduled.hatched == 1


def test_OpenOil_seeding():
    """Make sure special seed parameters comes through"""

    m = OpenDriftModel(
        drift_model="OpenOil",
        lon=-151,
        lat=60,
        do3D=True,
        start_time="2021-01-01T00:00:00",
        use_auto_landmask=True,
        m3_per_hour=5,
        droplet_diameter_max_subsea=0.1,
        droplet_diameter_min_subsea=0.01,
        droplet_diameter_mu=0.01,
        droplet_size_distribution="normal",
        droplet_diameter_sigma=10,
        oil_film_thickness=5,
        oil_type="GENERIC DIESEL",
    )

    m.o.set_config("environment:constant:x_wind", -1)
    m.o.set_config("environment:constant:y_wind", -1)
    m.o.set_config("environment:constant:x_sea_water_velocity", -1)
    m.o.set_config("environment:constant:y_sea_water_velocity", -1)
    m.seed()

    # to check impact of m3_per_hour: mass_oil for m3_per_hour of 1 * 5
    assert np.allclose(m.o.elements_scheduled.mass_oil, 8.412 * 5)
    assert m.o._config["m3_per_hour"]["value"] == 5
    assert m.o._config["droplet_diameter_max_subsea"]["value"] == 0.1
    assert m.o._config["droplet_diameter_min_subsea"]["value"] == 0.01
    assert m.o._config["droplet_diameter_mu"]["value"] == 0.01
    assert m.o._config["droplet_size_distribution"]["value"] == "normal"
    assert m.o._config["droplet_diameter_sigma"]["value"] == 10
    assert m.o.elements_scheduled.oil_film_thickness == 5
    assert m.o._config["oil_type"]["value"] == "GENERIC DIESEL"


def test_wind_drift():
    """Make sure changed wind drift numbers comes through"""

    m = OpenDriftModel(
        drift_model="OceanDrift",
        lon=-151,
        lat=60,
        do3D=True,
        wind_drift_factor=1,
        wind_drift_depth=10,
        start_time="2021-01-01T00:00:00",
        use_auto_landmask=True,
    )
    m.seed()
    assert m.o.elements_scheduled.wind_drift_factor == 1
    assert m.o._config["wind_drift_depth"]["value"] == 10


if __name__ == "__main__":
    unittest.main()



"""Test CLI methods."""

import os

from datetime import datetime

import pytest

import particle_tracking_manager as ptm


def test_setup():
    """Test CLI setup

    No drifters are run due to oceanmodel=None
    """
    ret_value = os.system(
        f"python {ptm.__path__[0]}/cli.py ocean_model=test lon=-151 lat=59 use_auto_landmask=True start_time='2000-1-1'"
    )
    assert ret_value == 0


def test_setup_library():
    """Same test but with library"""

    m = ptm.OpenDriftModel(
        ocean_model="test",
        lon=-151,
        lat=59,
        use_auto_landmask=True,
        steps=3,
        start_time=datetime(2000, 1, 1),
    )
    m.run_all()



"""Contains logic for configuring particle tracking simulations."""


# from docstring_inheritance import NumpyDocstringInheritanceMeta
import datetime
import json
import logging
import pathlib

from typing import Optional, Union

import pandas as pd

from .cli import is_None


_KNOWN_MODELS = [
    "NWGOA",
    "CIOFS",
    "CIOFSOP",
]

# Read PTM configuration information

loc = pathlib.Path(__file__).parent / pathlib.Path("the_manager_config.json")
with open(loc, "r") as f:
    # Load the JSON file into a Python object
    config_ptm = json.load(f)

# convert "None"s to Nones
for key in config_ptm.keys():
    if is_None(config_ptm[key]["default"]):
        config_ptm[key]["default"] = None


ciofs_operational_start_time = datetime.datetime(2021, 8, 31, 19, 0, 0)
ciofs_operational_end_time = (pd.Timestamp.now() + pd.Timedelta("48H")).to_pydatetime()
ciofs_end_time = datetime.datetime(2023, 1, 1, 0, 0, 0)
nwgoa_end_time = datetime.datetime(2009, 1, 1, 0, 0, 0)
overall_start_time = datetime.datetime(1999, 1, 1, 0, 0, 0)
overall_end_time = ciofs_operational_end_time


class ParticleTrackingManager:
    """Manager class that controls particle tracking model.

    Parameters
    ----------
    model : str
        Name of Lagrangian model package to use for drifter tracking. Only option
        currently is "opendrift".
    lon : Optional[Union[int,float]], optional
        Longitude of center of initial drifter locations, by default None. Use with `seed_flag="elements"`.
    lat : Optional[Union[int,float]], optional
        Latitude of center of initial drifter locations, by default None. Use with `seed_flag="elements"`.
    geojson : Optional[dict], optional
        GeoJSON object defining polygon for seeding drifters, by default None. Use with `seed_flag="geojson"`.
    seed_flag : str, optional
        Flag for seeding drifters. Options are "elements", "geojson". Default is "elements".
    z : Union[int,float], optional
        Depth of initial drifter locations, by default 0 but taken from the
        default in the model. Values are overridden if
        ``surface_only==True`` to 0 and to the seabed if ``seed_seafloor`` is True.
    seed_seafloor : bool, optional
        Set to True to seed drifters vertically at the seabed, default is False. If True
        then value of z is set to None and ignored.
    number : int
        Number of drifters to simulate. Default is 100.
    start_time : Optional[str,datetime.datetime,pd.Timestamp], optional
        Start time of simulation, by default None
    run_forward : bool, optional
        True to run forward in time, False to run backward, by default True
    time_step : int, optional
        Time step in seconds, options >0, <86400 (1 day in seconds), by default 300.
    time_step_output : int, Timedelta, optional
        How often to output model output. Should be a multiple of time_step.
        By default 3600.
    steps : int, optional
        Number of time steps to run in simulation. Options >0.
        steps, end_time, or duration must be input by user. By default steps is 3 and
        duration and end_time are None. Only one of steps, end_time, or duration can be
        non-None at initialization time. If one of steps, end_time, or duration is input
        later, it will be used to overwrite the three parameters according to that newest
        parameter.
    duration : Optional[datetime.timedelta], optional
        Length of simulation to run, as positive-valued timedelta object, in hours,
        such as ``timedelta(hours=48)``.
        steps, end_time, or duration must be input by user. By default steps is 3 and
        duration and end_time are None. For CLI, input duration as a pandas Timedelta
        string like "48h" for 48 hours. Only one of steps, end_time, or duration can be
        non-None at initialization time. If one of steps, end_time, or duration is input
        later, it will be used to overwrite the three parameters according to that newest
        parameter.

    end_time : Optional[datetime], optional
        Datetime at which to end simulation, as positive-valued datetime object.
        steps, end_time, or duration must be input by user. By default steps is 3 and
        duration and end_time are None. Only one of steps, end_time, or duration can be
        non-None at initialization time. If one of steps, end_time, or duration is input
        later, it will be used to overwrite the three parameters according to that newest
        parameter.

    ocean_model : Optional[str], optional
        Name of ocean model to use for driving drifter simulation, by default None.
        Use None for testing and set up. Otherwise input a string.
        Options are: "NWGOA", "CIOFS", "CIOFSOP".
        Alternatively keep as None and set up a separate reader (see example in docs).
    ocean_model_local : Optional, bool
        Set to True to use local version of known `ocean_model` instead of remote version.
    surface_only : bool, optional
        Set to True to keep drifters at the surface, by default None.
        If this flag is set to not-None, it overrides do3D to False, vertical_mixing to False,
        and the z value(s) 0.
        If True, this flag also turns off reading model output below 0.5m if
        drift_model is not Leeway:
        ``o.set_config('drift:truncate_ocean_model_below_m', 0.5)`` to save time.
    do3D : bool, optional
        Set to True to run drifters in 3D, by default False. This is overridden if
        ``surface_only==True``. If True, vertical advection and mixing are turned on with
        options for setting ``diffusivitymodel``, ``background_diffusivity``,
        ``ocean_mixed_layer_thickness``, ``vertical_mixing_timestep``. If False,
        vertical motion is disabled.
    vertical_mixing : bool, optional
        Set to True to include vertical mixing, by default False. This is overridden if
        ``surface_only==True``.
    use_static_masks : bool, optional
        Set to True to use static masks ocean_model output when ROMS wetdry masks are available, by default False.
        This is relevant for all of the available known models. If you want to use static masks
        with a user-input ocean_model, you can drop the wetdry_mask_rho etc variables from the
        dataset before inputting to PTM. Setting this to True may save computation time but
        will be less accurate, especially in the tidal flat regions of the model.
    output_file : Optional[str], optional
        Name of output file to save, by default None. If None, default is set in the model.

    Notes
    -----
    Configuration happens at initialization time for the child model. There is currently
    no separate configuration step.
    """

    logger: logging.Logger
    ocean_model: str
    lon: Union[int, float]
    lat: Union[int, float]
    surface_only: Optional[bool]
    z: Optional[Union[int, float]]
    start_time: Optional[Union[str, datetime.datetime, pd.Timestamp]]
    steps: Optional[int]
    time_step: int
    duration: Optional[datetime.timedelta]
    end_time: Optional[datetime.datetime]
    timedir: int
    config_ptm: dict
    config_model: Optional[dict]
    seed_seafloor: bool

    def __init__(
        self,
        model: str,
        lon: Optional[Union[int, float]] = None,
        lat: Optional[Union[int, float]] = None,
        geojson: Optional[dict] = None,
        seed_flag: str = config_ptm["seed_flag"]["default"],
        z: Optional[Union[int, float]] = config_ptm["z"]["default"],
        seed_seafloor: bool = config_ptm["seed_seafloor"]["default"],
        number: int = config_ptm["number"]["default"],
        start_time: Optional[Union[str, datetime.datetime, pd.Timestamp]] = None,
        run_forward: bool = config_ptm["run_forward"]["default"],
        time_step: int = config_ptm["time_step"]["default"],
        time_step_output: Optional[int] = config_ptm["time_step_output"]["default"],
        steps: Optional[int] = config_ptm["steps"]["default"],
        duration: Optional[datetime.timedelta] = config_ptm["duration"]["default"],
        end_time: Optional[datetime.datetime] = config_ptm["end_time"]["default"],
        # universal inputs
        ocean_model: Optional[str] = config_ptm["ocean_model"]["default"],
        ocean_model_local: Optional[bool] = config_ptm["ocean_model_local"]["default"],
        surface_only: Optional[bool] = config_ptm["surface_only"]["default"],
        do3D: bool = config_ptm["do3D"]["default"],
        vertical_mixing: bool = config_ptm["vertical_mixing"]["default"],
        use_static_masks: bool = config_ptm["use_static_masks"]["default"],
        output_file: Optional[str] = config_ptm["output_file"]["default"],
        **kw,
    ) -> None:
        """Inputs necessary for any particle tracking."""

        # get all named parameters input to ParticleTrackingManager class
        from inspect import signature

        sig = signature(ParticleTrackingManager)

        self.__dict__["config_ptm"] = config_ptm
        # self.__dict__["config_model"] = None  # previously defined before manager is initialized
        self.__dict__["_config_orig"] = None

        # check this here for initialization since later they will be set
        if steps is not None:
            assert duration is None and end_time is None
        if duration is not None:
            assert steps is None and end_time is None
        if end_time is not None:
            assert steps is None and duration is None

        # initialize all class attributes to None without triggering the __setattr__ method
        # which does a bunch more stuff
        for key in sig.parameters.keys():
            self.__dict__[key] = None

        # mode flags
        self.__dict__["has_added_reader"] = False
        self.__dict__["has_run_seeding"] = False
        self.__dict__["has_run"] = False

        # Set all attributes which will trigger some checks and changes in __setattr__
        # these will also update "value" in the config dict
        for key in sig.parameters.keys():
            # no need to run through for init if value is None (already set to None)
            if locals()[key] is not None:
                self.__setattr__(key, locals()[key])

        self.kw = kw

    def __setattr_model__(self, name: str, value) -> None:
        """Implement this in model class to add specific __setattr__ there too."""
        pass

    # calculate other simulation-length parameters when one is input
    # this way whichever parameter is input last overwrites the other parameters
    # that could have been input earlier
    # also have to check for the special case that start_time is being updated to be the
    # initial model output when the reader is set and in that case also need to update
    # end_time based on whichever of steps or duration is available.
    def calc_end_time(self, changed_variable):
        """Calculate end time based on other simulation length parameters."""

        if changed_variable == "steps" or (
            self.steps is not None and changed_variable == "start_time"
        ):
            return self.start_time + self.timedir * self.steps * datetime.timedelta(
                seconds=self.time_step
            )
        elif changed_variable == "duration" or (
            self.duration is not None and changed_variable == "start_time"
        ):
            return self.start_time + self.timedir * self.duration
        else:
            return self.end_time
        # if self.start_time is not None and self.steps is not None:
        #     return self.start_time + self.timedir * self.steps * datetime.timedelta(
        #         seconds=self.time_step
        #     )
        # elif self.start_time is not None and self.duration is not None:
        #     return self.start_time + self.timedir * self.duration
        # else:
        #     return self.end_time

    def calc_duration(self):
        """Calculate duration based on end_time and start_time."""
        if self.end_time is not None and self.start_time is not None:
            return abs(self.end_time - self.start_time)
        else:
            return self.duration

    def calc_steps(self):
        """Calculate steps based on duration and time_step."""
        if self.duration is not None and self.start_time is not None:
            return self.duration / datetime.timedelta(seconds=self.time_step)
        else:
            return self.steps

    def __setattr__(self, name: str, value) -> None:
        """Implement my own __setattr__ to enforce subsequent actions."""

        # create/update class attribute
        self.__dict__[name] = value

        # create/update "value" keyword in config to keep it up to date
        if name in self.config_ptm.keys():
            self.config_ptm[name]["value"] = value

        # create/update "value" keyword in model config to keep it up to date
        if self.config_model is not None:  # can't run this until init in model class
            self.__setattr_model__(name, value)

        # None of the following checks occur if value is None
        if value is not None:

            # check longitude when it is set
            if name == "lon":
                assert (
                    -180 <= value <= 180
                ), "Longitude needs to be between -180 and 180 degrees."

            if name == "lat":
                assert (
                    -90 <= value <= 90
                ), "Latitude needs to be between -90 and 90 degrees."

            if name == "start_time":
                if isinstance(value, (str, datetime.datetime, pd.Timestamp)):
                    self.__dict__[name] = pd.Timestamp(value)
                    self.config_ptm[name]["value"] = pd.Timestamp(value)
                else:
                    raise TypeError(
                        "start_time must be a string, datetime, or Timestamp."
                    )

            # # make sure ocean_model name uppercase
            # if name == "ocean_model":
            #     self.__dict__[name] = value.upper()
            #     self.config_ptm["ocean_model"]["value"] = value.upper()

            # check start_time relative to ocean_model selection
            if name in ["ocean_model", "start_time"]:
                if self.start_time is not None and self.ocean_model is not None:
                    assert isinstance(self.start_time, pd.Timestamp)
                    if self.ocean_model == "NWGOA":
                        assert overall_start_time <= self.start_time <= nwgoa_end_time
                    elif self.ocean_model == "CIOFS":
                        assert overall_start_time <= self.start_time <= ciofs_end_time
                    elif self.ocean_model == "CIOFSOP":
                        assert (
                            ciofs_operational_start_time
                            <= self.start_time
                            <= ciofs_operational_end_time
                        )

            # deal with if input longitudes need to be shifted due to model
            if name == "oceanmodel_lon0_360" and value:
                if self.ocean_model != "test" and self.lon is not None:
                    # move longitude to be 0 to 360 for this model
                    # this is not a user-defined option
                    if -180 < self.lon < 0:
                        self.__dict__["lon"] += 360
                        self.config_ptm["lon"]["value"] += 360  # this isn't really used

            if name == "surface_only" and value:
                self.logger.info(
                    "Overriding values for do3D, z, and vertical_mixing because surface_only is True (to False, 0, False)."
                )
                self.do3D = False
                self.z = 0
                self.vertical_mixing = False

            # in case any of these are reset by user after surface_only is already set
            # if surface_only is True, do3D must be False
            if name == "do3D" and value and self.surface_only:
                self.logger.info(
                    "do3D must be False because surface_only is True. Setting do3D to False."
                )
                self.__dict__["do3D"] = False
                self.config_ptm["do3D"]["value"] = False

            # if surface_only is True, z must be 0
            if name == "z" and value != 0 and self.surface_only:
                self.logger.info(
                    "z must be 0 because surface_only is True. Setting z to 0."
                )
                self.__dict__["z"] = 0
                self.config_ptm["z"]["value"] = 0

            # if surface_only is True, vertical_mixing must be False
            if name == "vertical_mixing" and value and self.surface_only:
                self.logger.info(
                    "vertical_mixing must be False because surface_only is True. Setting vertical_mixing to False."
                )
                self.__dict__["vertical_mixing"] = False
                self.config_ptm["vertical_mixing"]["value"] = False

            # if not 3D turn off vertical_mixing
            if name in ["do3D", "vertical_mixing"]:
                if not self.do3D and self.vertical_mixing:
                    self.logger.info("Turning off vertical_mixing since do3D is False")
                    self.__dict__["vertical_mixing"] = False
                    self.config_ptm["vertical_mixing"]["value"] = False

            # set z to None if seed_seafloor is True
            if name == "seed_seafloor" and value:
                self.logger.info("setting z to None since being seeded at seafloor")
                self.z = None

            # in case z is changed back after initialization
            if (
                name == "z" and self.seed_seafloor
            ):  # already checked that value is not None
                self.logger.info(
                    "setting `seed_seafloor` from True to False since now setting a non-None z value"
                )
                self.seed_seafloor = False

            # if reader, lon, and lat set, check inputs
            if (
                name == "has_added_reader"
                and value
                and self.lon is not None
                and self.lat is not None
                or name in ["lon", "lat"]
                and self.has_added_reader
                and self.lon is not None
                and self.lat is not None
            ):

                if self.ocean_model != "test":
                    rlon = self.reader_metadata("lon")
                    assert rlon.min() < self.lon < rlon.max()
                    rlat = self.reader_metadata("lat")
                    assert rlat.min() < self.lat < rlat.max()

            # if reader, lon, and lat set, check inputs
            if name == "has_added_reader" and value and self.start_time is not None:

                if self.ocean_model != "test":
                    assert self.reader_metadata("start_time") <= self.start_time

            # if reader, lon, and lat set, check inputs
            if name == "has_added_reader" and value:
                assert self.ocean_model is not None

            # define time direction
            if name == "run_forward":
                if value:
                    self.__dict__["timedir"] = 1
                else:
                    self.__dict__["timedir"] = -1

            if (
                name in ["start_time", "end_time", "steps", "duration"]
                and self.start_time is not None
            ):
                # the behavior in calc_end_time changes depending on which variable has been updated
                self.__dict__["end_time"] = self.calc_end_time(name)
                self.config_ptm["end_time"]["value"] = self.calc_end_time(name)
                # duration and steps are always updated now that start_time and end_time are set
                self.__dict__["duration"] = self.calc_duration()
                self.config_ptm["duration"]["value"] = self.calc_duration()
                self.__dict__["steps"] = self.calc_steps()
                self.config_ptm["steps"]["value"] = self.calc_steps()

            if name == "ocean_model" and value not in _KNOWN_MODELS:
                self.logger.info(f"ocean_model is not one of {_KNOWN_MODELS}.")

    def add_reader(self, **kwargs):
        """Here is where the model output is opened."""

        self.run_add_reader(**kwargs)

        self.has_added_reader = True

    def seed(self, lon=None, lat=None, z=None):
        """Initialize the drifters in space and time

        ... and with any special properties.
        """

        for key in [lon, lat, z]:
            if key is not None:
                self.__setattr__(self, f"{key}", key)

        # if self.ocean_model != "test" and not self.has_added_reader:
        #     raise ValueError("first add reader with `manager.add_reader(**kwargs)`.")

        # have this check here so that all parameters aren't required until seeding
        if self.seed_flag == "elements" and self.lon is None and self.lat is None:
            msg = f"""lon and lat need non-None values if using `seed_flag="elements"`.
                    Update them with e.g. `self.lon=-151` or input to `seed`."""
            raise KeyError(msg)
        elif self.seed_flag == "geojson" and self.geojson is None:
            msg = f"""geojson need non-None value if using `seed_flag="geojson"`."""
            raise KeyError(msg)

        msg = f"""z needs a non-None value.
                  Please update it with e.g. `self.z=-10` or input to `seed`."""
        if not self.seed_seafloor:
            assert self.z is not None, msg

        if self.ocean_model is not None and not self.has_added_reader:
            self.add_reader()

        if self.start_time is None:
            raise KeyError(
                "first add reader with `manager.add_reader(**kwargs)` or input a start_time."
            )

        self.run_seed()
        self.has_run_seeding = True

    def run(self):
        """Call model run function."""

        if not self.has_run_seeding:
            raise KeyError("first run seeding with `manager.seed()`.")

        self.logger.info(
            f"start_time: {self.start_time}, end_time: {self.end_time}, steps: {self.steps}, duration: {self.duration}"
        )

        # need end time info
        assert (
            self.steps is not None
            or self.duration is not None
            or self.end_time is not None
        )

        self.run_drifters()
        self.has_run = True

    def run_all(self):
        """Run all steps."""

        if not self.has_added_reader:
            self.add_reader()

        if not self.has_run_seeding:
            self.seed()

        if not self.has_run:
            self.run()

    def output(self):
        """Hold for future output function."""
        pass

    def _config(self):
        """Model should have its own version which returns variable config"""
        pass

    def _add_ptm_config(self):
        """Have this in the model class to modify config"""
        pass

    def _add_model_config(self):
        """Have this in the model class to modify config"""
        pass

    def _update_config(self) -> None:
        """Update configuration between model, PTM additions, and model additions."""

        # Modify config with PTM config
        self._add_ptm_config()

        # Modify config with model-specific config
        self._add_model_config()

    def show_config_model(self):
        """define in child class"""
        pass

    def show_config(self, **kwargs) -> dict:
        """Show parameter configuration across both model and PTM."""

        self._update_config()

        # Filter config
        config = self.show_config_model(**kwargs)

        return config

    def reader_metadata(self, key):
        """define in child class"""
        pass

    def query_reader(self):
        """define in child class."""
        pass

    def all_export_variables(self):
        """Output list of all possible export variables.

        define in child class.
        """
        pass

    def export_variables(self):
        """Output list of all actual export variables.

        define in child class.
        """
        pass

    @property
    def outfile_name(self):
        """Output file name.

        define in child class.
        """
        pass



from setuptools import setup


setup(
    use_scm_version={
        "write_to": "particle_tracking_manager/_version.py",
        "write_to_template": '__version__ = "{version}"',
        "tag_regex": r"^(?P<prefix>v)?(?P<version>[^\+]+)(?P<suffix>.*)?$",
    },
    entry_points={"console_scripts": ["ptm=particle_tracking_manager.cli:main"]},
)
